/*--------------------------------------------------------*/
/* Programmed by OZ						                  */
/* File Name : supermarket_source.cpp                     */
/*--------------------------------------------------------*/

#include <stdio.h>   // printf()
#include <malloc.h>  // malloc()
#include <time.h>    // time_t structure
#include <stdlib.h>  // srand()
#include <conio.h>   // getch() (for testing)

// #define INTERVAL_MAX 4
#define CUSTOMER_MAX 10001 // the range of customer's name (0 ~ 10000)

/* Queue Node Structure (I used a double linked list to implement a adaptable queue) */
typedef struct _qnode
{
	int waitTime; // customer's arrival time, It means the period of interval per customer
	struct _qnode *prev; // previous queue node pointer
	struct _qnode *next; // next queue node pointer
} qnode; // typedef

/* Binary Tree Node Structure */
typedef struct _tnode
{
	int customerNo; // customer's number that is given by computer randomly
	struct _tnode *left; // tree's left node pointer
	struct _tnode *right; // tree's right node pointer
} tnode; // typedef

/* Variables for Queue */
qnode *head, *tail; // Pointer variables for queue (head pointer & tail pointer)
int queueDataCount; // the number of data in queue
int maxNumber = 0; // the maximum number of customers in the queue at particular time
int maxTime; // the time when maximum number of customers are in the queue
int maxWaitTime = 0; // the longest waiting time of a customer among all customers

/* Variables for Binary Tree */
tnode *base; // the base node of tree structure
bool customerNo[10001]; // 0 ~ 10000
int curCustomerNo; // current customer's name(number)

/* Functions for Queue */
void InitQueue(); // Initialize queue
int Enqueue(int value); // Put data to rear of queue
int Dequeue(); // Get data from front of queue
void ClearQueue(); // Release all queue data
void AddWaitTimeToQueue(); // Add the waiting time at the line per customer (to get longest wait time)

/* Functions for Binary Tree */
void InitTree(int initValue); // Initialize tree
tnode* InsertToTree(int value); // Insert data to proper position of tree
int GetNewCustomerNumber(); // Get new customer's name(number) without overlapping (Get unique customer's name)
void OutputTree(tnode *node, int totalSpace, int depth); // Output the tree on screen (I refered the Exercise 17.25 in the textbook for the printing algorithm)

// Main Function
void main(int argc, char **argv) // Extra paramenter (max-arrival-interval, max-service-interval, simulate-time)
{
	int intervalMaxArrive = 4; // The maximum interval value of arrival time, default : 1 ~ 4 minutes
	int intervalMaxService = 4; // The maximum inerval value of service time, default : 1 ~ 4 minutes
	int simulatePeriod = 720; // Simulation time, default : 720 minutes

	if (argc == 4) // If user input the additional parameter
	{
		intervalMaxArrive = atoi(argv[1]); // Set new interval value of arrival time
		intervalMaxService = atoi(argv[2]); // Set new interval value of arrival time
		simulatePeriod = atoi(argv[3]);
	}

	time_t t; // time_t structure (To make real random number every time
	srand(time(&t)); // Set current time to a random starting point 
	int count = 0; // current time in while loop, loop variable
    int waitTime; // Save the each customer's waiting time, I used this variable to find the value of 'maxWaitTime'
	int arrivalTime; // Save arrival time of current customer
	int serviceTime; // Save service time of current customer
	
	InitQueue(); // Initialize Queue
	InitTree(0); // Initialize Tree

	// Initialize the 'customerNo' array (This array will be used to give the name to custmers
	for(int i = 0; i < CUSTOMER_MAX; i++)
		customerNo[i] = true; // At the first stage, all numbers(0 ~ 10000) are available

	arrivalTime = rand() % intervalMaxArrive + 1; // Get first customer's arrival time
	count += arrivalTime; // 'count' minutes are passed (After 'count' minutes, the first customer comes to the checkout line

	curCustomerNo = GetNewCustomerNumber(); // Get the first customer's name(number)

	InsertToTree(curCustomerNo); // Insert first customer's name to tree
	customerNo[curCustomerNo] = false; // Now, we can't use the first customer's name (Because of condition that customer's name is unique for every customer)

	serviceTime = rand() % intervalMaxService + 1; // Generate first customer's srervice time
	arrivalTime = rand() % intervalMaxArrive + 1; // Generate the interval time between first customer's arrival time and second customer's arrival time
	
	while(count < simulatePeriod) // Loop ( 0 ~ 'simulatePeriod' ) default value of 'simulatePeriod' is 720(minutes)
	{
		if(maxNumber < queueDataCount) // If the current data number in queue is larger than 'maxNumber'
		{
			maxNumber = queueDataCount; // Set new data to 'maxNumber' ('maxNumber' is the maximum number of customers in the queue at particular time)
			maxTime = count; // // Update 'maxTime' ('maxTime' is the time when maximum number of customers are in the queue)
		}

		arrivalTime--; // Decrease the 'arrivalTime' (time passes) 

		if(arrivalTime == 0) // If the 'arrivalTime' is zero (It means that new customer comes to the checkout line)
		{
			printf(" in "); // Print 'in' on screen
			Enqueue(0); // Put zero to queue (enqueue, the 0 means the initial waiting time of each customer. because customer that comes to the checkout line right now didn't wait at all)
			InsertToTree(GetNewCustomerNumber()); // Get new unique customer's name(number) and input it to tree
			arrivalTime = rand() % intervalMaxArrive + 1; // Generate new arrival interval time
		}

		serviceTime--; // Decrease the 'serviceTime' (time passes)

		if(serviceTime == 0) // If the 'serviceTime' is zero (It means that service for previous customer is ended)
		{
			printf(" out "); // Print 'out' on screen
			waitTime = Dequeue(); // Get waiting time of each customers (dequeue)
			serviceTime = rand() % intervalMaxService + 1; // Generate new service interval time

			if(maxWaitTime < waitTime) // If waiting time of current customer who finish checkout right now is larger than 'maxWaitTime'
			{
				maxWaitTime = waitTime; // Set new data to 'maxWaitTime'
			}
		}

		AddWaitTimeToQueue(); // Compute the waiting time of all customers in queue (Whenever 1 minute passes, the 1(minute) is added to the customers' waiting time

		count++; // time passes (1 minute passes)

		printf("\nNumber of Customers in the line : %2d, Current Time : %3d", queueDataCount, count); // Output the result(customer's in or out information at this time)
	}

	printf("\n\nMaximum Number : %d customers at %d minute", maxNumber, maxTime); // Output the maximum number of customers in the queue at particular time
	printf("\nMaximum Waiting Time : %d minuts", maxWaitTime); // Output the longest wait time of particular customer among all customers
	printf("\n\n");

	OutputTree(base->left, 0, 0); // Output the tree structure (by 5 levels)

	ClearQueue(); // Release Queue data
}

/*--------------------------------------------------------*/
/* The functions for QUEUE                                */
/*--------------------------------------------------------*/
/* Initialize queue */
void InitQueue()
{
	head = (qnode *)malloc(sizeof(qnode)); // allocate new memory to head node of queue
	tail = (qnode *)malloc(sizeof(qnode)); // allocate new memory to tail node of queue
	head->prev = head; // head->prev point to head
	head->next = tail; // head->next point to tail
	tail->prev = head; // tail->prev point to head
	tail->next = tail; // tail->next point to tail

	queueDataCount = 0; // Initialize the queueDataCount (Empty queuq)
}

/* Put data to rear of queue */
int Enqueue(int value)
{
	qnode *newNode; // queue node pointer

	if((newNode = (qnode *)malloc(sizeof(qnode))) == NULL) // memory allocation for new node of queue
	{
		printf("\nOut of memory."); // fail to allocate new memory
		return -1;
	}

	newNode->waitTime = value; // set data of new node
	tail->prev->next = newNode; // insert 'newNode' as previous node of tail
	newNode->prev = tail->prev; // newNode->prev point tail->prev
	tail->prev = newNode; // tail->prev point to newNode
	newNode->next = tail; // newNode->next point to tail

	queueDataCount++; // Increase the data number in queue

	return value;
}

/* Get data from front of queue */
int Dequeue() 
{
	qnode *delNode; // queue node pointer
	delNode = head->next; // Initialize delNode

	if(delNode == tail) // If delNode is tail node
	{
		printf("\nNo Customers"); // Underflow
		return -1;
	}

	int i = delNode->waitTime; // Get the customer's waiting time (this customer's data will be deleted in the queue)
	head->next = delNode->next; // head->next point to delNode->next
	delNode->next->prev = head; // delNode->next->prev point to head
	free(delNode); // Release delNode

	queueDataCount--; // Decrease the data number in queue

	return i; // return customer's waiting time
}

/* Release all queue data */
void ClearQueue() 
{
	qnode *node1; // queue node pointer 1
	qnode *node2; // queue node pointer 2
	node1 = head->next; // Initialize node1

	while(node1 != tail) // While node1 is not tail node
	{   // Step by step release of memory for queue nodes
		node2 = node1;
		node1 = node1->next;
		free(node2); // release node2
	}

	head->next = tail; // Empty queue
	tail->prev = head;
}

/* Add the waiting time at the line per customer */
void AddWaitTimeToQueue()
{
	qnode *curNode; // queue node pointer1
	curNode = head->next; // Initialize curNode

	while(curNode != tail) // While curNode is not tail node
	{
		curNode->waitTime++; // Add 1 to waitTime of all customers
		curNode = curNode->next; // Move to next node
	}
}

/*--------------------------------------------------------*/
/* The functions for TREE                                 */
/*--------------------------------------------------------*/
/* Initialize tree */
void InitTree(int initValue)
{
	base = (tnode *)malloc(sizeof(qnode)); // memory allocation for base node of tree
	base->customerNo = initValue; // Initialize base node's data(customer name), but I didn't use the base node's data as real data
	base->left = NULL; // Initialize base->left node
	base->right = NULL; // Initialize base->right node
}

/* Insert data to proper position of tree */
tnode* InsertToTree(int value)
{
	tnode *node1, *node2; // two tree node pointers
	node1 = base; // node1 is parent node of node2
	node2 = base->left; // node2 is left-child node of node1

	while(node2 != NULL) // While node2 is not null
	{
		node1 = node2; // Maintain the parent node

		if(value < node2->customerNo) // If new data is smaller than node2's data
			node2 = node2->left; // Go to left direction of tree
		else
			node2 = node2->right; // Go to right direction of tree
	} // If the while loop ends, We can get the proper position to insert new data

	if((node2 = (tnode *)malloc(sizeof(tnode))) == NULL) // Memory allocation for new data
		return NULL;

	node2->customerNo = value; // Assign new data to new node
	node2->left = NULL; // Initialize left and right pointer node of node2
	node2->right = NULL;

	if(value < node1->customerNo || node1 == base) // If node2 is smaller than node1 node or if node1 is base node
		node1->left = node2; // Take the node2 as node1's left-child node
	else
		node1->right = node2; // Take the node2 as node1's right-child node

	return node2; // return new node inserted to tree
}

/* Get new customer's name without overlapping */
int GetNewCustomerNumber()
{
	int number; // Temporary data to save the current node's data

	while(1) // Infinite loop
	{
		number = rand() % CUSTOMER_MAX; // Generate new customer's name

		// If the name that is generated is arleady in use, Generate new customer's name again
		if(customerNo[number] == true) // If the name is available
		{
			customerNo[number] = false; // Now, this name can't be used any more
			break; // Escape the loop
		}
	}

	return number; // Return the number is generated
}

/* Output the tree on screen */
/* node : node pointer that is worked currently */
/* totalSpace : Represent the number of spaces preceding the value to be output */
/* depth : node's level position */
/* Recursive Function : Tree Traverse - Inorder	Traverse Process */
void OutputTree(tnode *node, int totalSpace, int depth)
{
	if(depth >= 5) // If current node's position is deeper than level 5 of tree, return
		return;

	if(node == NULL && depth < 5) // If current node is null and the node's level is lower than 5
		printf("\n"); // Move to next line

	if(node != NULL) // If current node is not null
	{
		OutputTree(node->right, totalSpace+5, depth + 1); // Recursively call OutputTree (right node)

		for(int i = 0; i < totalSpace; i++) // For loop - 0 ~ totalSpace, Make empty space (Show tree sturcture on screen)
			printf("  ");

		printf("%5d\n", node->customerNo); // Output current data(customer's name)

		OutputTree(node->left, totalSpace+5, depth + 1); // Recursively call OutputTree (left node)
	}
}